#===================================================
#
# Project      : mlsuite model prediction
# Description  : predict testing dataset by 13 models
# Usage        : 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-06-14
#
#====================================================


import joblib
import os
import re
import glob
import time
from .MLOpenWrite import OpenM, Openf
from .MLSupervising import HyperparamOptimal, MergeAllResult

class Supervised():
    def __init__(self, arg, log, Type='C', MClass=2):
        self.arg = arg
        self.log = log
        self.Type=Type
        self.MClass= MClass

    def GetLatest(self, _Y_name):
        comples = '%s%s_Class_best_estimator_' % (self.arg.MODELF, _Y_name) 
        modelfs = glob.glob(comples + '*.pkl')
        modelnu = sorted([ int(re.search('_best_estimator_(?P<id>\d+).pkl', i ).group('id')) for i in modelfs ])
        if self.arg.refit == 'raw':
            assign = 0
        elif self.arg.refit == 'fixed':
            assign = 1
        elif self.arg.refit == 'latest':
            assign = max(modelnu)
        else:
            assign = int(self.arg.refit)
        return '%s%s.pkl' % (comples, assign)

    def Predict_X(self, DFall, _Y_name, Xa):
        Best_Model = joblib.load(self.GetLatest(_Y_name))
        CVRlt = HyperparamOptimal(arg=self.arg, log=self.log, model=self.arg.model, y_variable_type=self.Type, y_variable_class=self.MClass)
        Y_PRED, Y_Eval = CVRlt.Predicting_C(Best_Model, DFall, _Y_name, Xa)
        ### All_Pred_predict
        merge_opt = MergeAllResult(log=self.log)
        merge_opt.CVMerge(Y_PRED, label='Predict')
        Openf('%s%s_Class_%s_detials.xls'%(self.arg.output, _Y_name, 'Predict'), (merge_opt.allTest), index=True, index_label='sample').openv()
        Openf('%s%s_Class_%s_summary.xls'%(self.arg.output, _Y_name, 'Predict'), (merge_opt.ALL_Result), index=True, index_label='sample').openv()
        Openf('%s%s_Class_%s_summary_statistics.xls'%(self.arg.output, _Y_name, 'Predict'), (merge_opt.EvalueA), index=True, index_label='Score').openv()
        ### All_evaluate_score_
        merge_opt.ModelScore(Y_Eval)
        Openf('%s%s_Class_%s_metrics_pvalues.xls'%(self.arg.output, _Y_name, 'Predict'), (merge_opt.All_pvalues), index=True, index_label='Score').openv()



def run_predict(args, Log):
    ### set predict output directory
    if not ( bool(args.out_predict) | bool(args.modelpath) ):
        args.output = '%s/02Prediction/%s' % (args.outdir, args.header)
        Log.NIF('')
        Log.NIF('')
    os.makedirs(os.path.dirname(args.output), exist_ok=True)

    ### load data
    (group, RYa, CYa, Xa, Xg) = OpenM(args, Log).openg()
    YType = group[(group.Variables.isin(RYa + CYa))][['Variables', 'Type']]
    Predt_set = OpenM(args, Log).openp()
    for Yi,Typi in YType.values:
        Log.CIF(('%s: Supervised Predicting' % Yi).center(45, '*'))
        Supervised(args, Log, Type=Typi, MClass=2).Predict_X(Predt_set, Yi, Xa)
        Log.CIF(('%s: Supervised Predicting Finish' % Yi).center(45, '*'))
    Log.CIF('Supervisor Predicting Finish'.center(45, '*'))


if __name__ == "__main__":
    from MLLogging import Logger
    from MLArguments import Args
    args = Args()
    args.commands = "Predict"
    args.group = "../Data/fragment.stat_Methy.Ratio.PCA_0.95.group.Info.PCA.txt"
    args.predict = "../Data/fragment.stat_Methy.Ratio.PCA_0.95.validation.PCA.txt"
    args.modelpath = "../Test"
    args.model = "LinearSVM"
    args.out_predict = "../Test"
    args.out_header = "Predict"
    args.refit = "fixed"
    args.n_job = 5

    # print output message and create new folder for modeling results
    args.header = '/%s_' % (args.model)
    args.outdir = '%s/%s/' % (args.out_predict, args.model)
    args.MODELF = '%s/00MODEL/%s' % (args.outdir, args.header)
    Theader = time.strftime("%Y%m%d%H%M", time.localtime())
    args.output = '%s/%s_Prediction_%s/%s' % (args.outdir, args.out_header, Theader, args.header)
    os.makedirs(os.path.dirname(args.output), exist_ok=True)
    Log = Logger('%s%s_log.log' % (args.output, args.commands))
    Log.NIF("The argument you have set as follows:".center(59, '*'))
    for i, k in enumerate(vars(args), start=1):
        Log.NIF('**%s|%-13s: %s' % (str(i).zfill(2), k, str(getattr(args, k))))
    Log.NIF(59 * '*')
    Log.CIF( 'Supervisor Predicting'.center(45, '*') )

    run_predict(args, Log)



#===================================================
#
# Project      : mlsuite building models or predicting
# Description  : argspase script
# Usage        : 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-06-21
#
#====================================================

import argparse
import os

def Args():
    parser = argparse.ArgumentParser(
                formatter_class=argparse.RawDescriptionHelpFormatter,
                prefix_chars='-+',
                conflict_handler='resolve',
                description="\nThe traditional machine learning analysis based on sklearn package:\n",
                epilog='''\
Example:
1. python mlsuite.py Auto    -i data.traintest.txt -g group.new.txt -p data.predict.txt -o testdt/ -m DT -pn 0.9 (The cumulative contribution ratio of PCA is 0.9.)
2. python mlsuite.py Auto    -i data.traintest.txt -g group.new.txt -p data.predict.txt -o testdt/ -m DT -pn 0.9 -f (10 times sample grouping and 10 models)
3. python mlsuite.py Auto    -i data.traintest.txt -g group.new.txt -o testdt/ -m DT -pn 0.9
4. python mlsuite.py Fitting -i data.traintest.txt -g group.new.txt -o testdt/ -m DT -pn 0.9
5. python mlsuite.py Fitting -i data.traintest.txt -g group.new.txt -o testdt/ -m DT -pn 0.9 -f (10 times sample grouping and 10 models)
6. python mlsuite.py Predict -p data.predict.txt   -g group.new.txt -x modelpath/ -y predictdt/ -m DT -pn 0.9.''')

    parser.add_argument('-V','--version',action ='version',
                version='mlsuite version 2.1.7')
    subparsers = parser.add_subparsers(dest="commands",
                    help='machine learning models help.')
    ### Fitting module
    P_fitting  = subparsers.add_parser('Fitting',conflict_handler='resolve', add_help=False)
    P_fitting.add_argument("-f", "--fit_10", action="store_true", default=False,
                    help='''fitting modeling by 10 times sample spliting''')
    P_fitting.add_argument("-i", "--input",type=str,
                    help='''the input train and test data file with dataframe format  by row(samples) x columns (features and Y). the sample column name must be Sample.
''')
    P_fitting.add_argument("-g", "--group",type=str,required=True,
                    help='''the group file tell the featues, groups and variable type, which has Variables, Group, Type columns. Only continuous and discrete variables are supported in variable type. Onehot variables is coming.''')
    P_fitting.add_argument("-o", "--outdir",type=str,default=os.getcwd(),
                    help="output file dir, default=current dir.")
    P_fitting.add_argument("-m", "--model",type=str, default='LinearSVM',
                    help='''the model you can used for ML.
You can choose the models as follows:
classification:......++++++++++++++++++++++
**RF.................RandomForestClassifier
**GBDT...............GradientBoostingClassifier
**XGB................XGBClassifier(+LR/LRCV)
**MLP................MLPClassifier
**DT.................DecisionTreeClassifier
**AdaB_DT............AdaBoostClassifier(DT)
**LinearSVM..........LinearSVC(penalty='l1')
**LinearSVMil2.......LinearSVC(penalty='l2')
**SVMlinear..........SVC(kernel="linear")
**SVM................SVC(no linear)
**nuSVMrbf...........NuSVC(kernel='rbf')
**SGD................SGDClassifier
**KNN................KNeighborsClassifier
**RNN................RadiusNeighborsClassifier
**MNB................MultinomialNB
**CNB................ComplementNB
**BNB................BernoulliNB
**GNB................GaussianNB
**LR.................LogisticRegression
**LRCV...............LogisticRegressionCV
''')
    P_fitting.add_argument("-am", "--Addmode", type=str, default='LinearSVM',
                    help='''use additional model to adjust modeling in RF, GBDT and XGB estimators. N means no addition.''')
    P_fitting.add_argument("-nt", "--n_iter", type=int, default= 2500,
                    help="Number of parameter settings that are sampled in RSCV. n_iter trades off runtime vs quality of the solution.")
    P_fitting.add_argument("-nj", "--n_job", type=int, default=-1,
                    help="Number of cores to run in parallel while fitting across folds.")
    P_fitting.add_argument("-cm", "--CVfit", type=str, default='SSA',
                    help='''the cross validation model:
**SSA..................StratifiedShuffleSplit() + StratifiedKFold()
**SSS..................StratifiedShuffleSplit()
**SKF..................StratifiedKFold()
**RSKF.................RepeatedStratifiedKFold()
**RKF..................RepeatedKFold()
**LPO..................LeavePOut()
**LOU..................LeaveOneOut()
''')
    P_fitting.add_argument("-pn", "--pc_number", type=float, default=0,
                    help='''select PC numbers, default is 0 that means not to do PCA analysis, >1 means selecting PC numbers,
                            >0 and <1 means select the number of components such that the amount of variance that needs to be
                            explained is greater than the percentage specified by pc_number.''')
    P_fitting.add_argument("-pc", "--pc_center", action="store_false",
                    help='''PCA input data center, default is not open. If open, parameter "pc_number" is not 0.''')
    P_fitting.add_argument("-ps", "--pc_scale", action="store_true",
                    help='''PCA input data scale, default is not open. If open, parameter "pc_number" is not 0.''')
    P_fitting.add_argument("-am", "--Addmode", type=str, default='LinearSVM',
                    help='''use additional model to adjust modeling in RF, GBDT and XGB estimators. N means no addition.''')
    P_fitting.add_argument("-rd", "--random",type=int, default=123456,
                    help="the random seeds for cross validation when using StratifiedShuffleSplit.")
    P_fitting.add_argument("-tz", "--testS",type=float, default=0.3,
                    help="the test size for cross validation when using StratifiedShuffleSplit.")
    P_fitting.add_argument("-sc", "--SearchCV",type=str, nargs='?', default='GSCV', choices=['GSCV','RSCV',],
                    help="the hyperparameters optimization method. You also can set it none to discard the method.")
    P_fitting.add_argument("-rt", "--Repeatime", type=int, default=1,
                    help='''the repeat time of modeling. Suggest change to a lagger value with a none value of SearchCV.''')
    P_fitting.add_argument("-cz", "--GStestS",type=float, default=0.3,
                    help="the test size for cross validation when using StratifiedShuffleSplit in gridsearchCV.")
    P_fitting.add_argument("-cv", "--crossV",type=int, default=10,
                    help="the cross validation times when using StratifiedShuffleSplit.")
    P_fitting.add_argument("-lp", "--leavP",type=int, default=1,
                    help="the size of the test sets in Leave-P-Out cross-validator.")
    P_fitting.add_argument("-cc", "--calib", action='store_true', default=True,
                    help='''whether use CalibratedClassifierCV to calibrate probability with isotonic regression or sigmoid.''')
    P_fitting.add_argument("-si", "--calibme", type=str, default='sigmoid', choices=['isotonic','sigmoid'],
                    help="the CalibratedClassifierCV method, you can choose isotonic and sigmoid.")
    P_Fitting  = subparsers.add_parser('Fitting',conflict_handler='resolve',
                    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                    parents=[P_fitting],
                    help='Fitting and predicting the training and testing set from estimators.')
    ### Predict module
    P_predict  = subparsers.add_parser('Predict', conflict_handler='resolve',add_help=False,)
    P_predict.add_argument("-p", "--predict",type=str,
                    help="the predict matrix file.")
    P_predict.add_argument("-x", "--modelpath",type=str,
                    help="the model path used for prediction.")
    P_predict.add_argument("-y", "--out_predict",type=str,
                    help="the predict result file path.")
    P_predict.add_argument("-z", "--out_header",type=str, default='Predict',
                    help="the predict result file header.")
    P_predict.add_argument("-ef", "--refit" , default='fixed',
                    help='''the model used for prediction.
raw: the train set of cross validation in modeling data.
fixed: the train and test set of cross validation in modeling data.
latest: the latest data of modeling data and prediction data.
int number: specified number model used for prediction ''')
    P_Predict  = subparsers.add_parser('Predict',conflict_handler='resolve',
                    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                    parents=[P_fitting, P_predict],
                    help='predict new data from fittting process model.')
    ### Auto module
    P_Autopipe = subparsers.add_parser('Auto', conflict_handler='resolve', prefix_chars='-+',
                    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                    parents=[P_fitting, P_predict],
                    help='the auto-processing for all: Fitting and/or Prediction.')

    args  = parser.parse_args()
    return args



import logging

class DispatchingFormatter:
    def __init__(self, formatters, default_formatter):
        self._formatters = formatters
        self._default_formatter = default_formatter

    def format(self, record):
        formatter = self._formatters.get(record.name, self._default_formatter)
        return formatter.format(record)

class Logger:
    level_dict = {
        'NOTSET'  : logging.NOTSET,
        'DEBUG'   : logging.DEBUG,
        'INFO'    : logging.INFO,
        'WARNING' : logging.WARNING,
        'ERROR'   : logging.ERROR,
        'CRITICAL': logging.CRITICAL,
    }

    ChangeFrom = DispatchingFormatter(
            { 'c' : logging.Formatter( '[%(asctime)s] [%(levelname)-4s]: %(message)s', '%Y-%m-%d %H:%M:%S'),
              'p' : logging.Formatter( '[%(levelname)-4s]: %(message)s'),
              'n' : logging.Formatter( '%(message)s' ),
            }, 
            logging.Formatter('%(message)s')
     )

    def __init__(self, outpath, filemode='w',  clevel = 'INFO', Flevel = 'INFO'):

        logging.basicConfig(
            level    = Logger.level_dict[clevel] ,
            format   = '[%(asctime)s] [%(levelname)-4s]: %(message)s',
            datefmt  = '%Y-%m-%d %H:%M:%S',
            filename = None,
        )

        File = logging.FileHandler(outpath,  mode= filemode)
        File.setLevel(Logger.level_dict[Flevel])
        File.setFormatter(Logger.ChangeFrom)
        logging.getLogger().addHandler(File)
        #Hand = logging.StreamHandler()
        #Hand.setFormatter(Logger.ChangeFrom)
        #logging.getLogger().addHandler(Hand)

        self.R = logging
        self.C = logging.getLogger('c')
        self.P = logging.getLogger('p')
        self.N = logging.getLogger('n')
        self.CIF = logging.getLogger('c').info
        self.NIF = logging.getLogger('n').info
        self.CWA = logging.getLogger('c').warning
        self.NWA = logging.getLogger('n').warning


#===================================================
#
# Project      : mlsuite build model
# Description  : build 13 models by StratifiedShuffleSplit
# Usage        : ./MLSupervising.py 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-05-12
#
#====================================================


from .MLUtilities import MyThread, _predict_proba_lr, CrossvalidationSplit, Check_Label, Check_Binar
from .MLEstimators import ML
from .MLOpenWrite import OpenM, Openf
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.metrics import (accuracy_score, f1_score, recall_score, precision_score,
                             balanced_accuracy_score, roc_auc_score, average_precision_score)
from sklearn.preprocessing import OneHotEncoder
from sklearn.calibration import CalibratedClassifierCV
from scipy.sparse import vstack
import os
import numpy as np
import pandas as pd
import re
import time
import joblib
pd.set_option('display.max_rows', 100000)
pd.set_option('display.max_columns', 100000)
pd.set_option('display.width', 100000)


class pca_svd():
    def __init__(self, k):
        self.k = k

    def fit(self, x, center=True, scale=False):
        '''
        x: pca降维初始数据， np.arary数组，行表示样本，列表示特征
        '''
        self.x = np.asarray(x)
        self.x_mean = self.x.mean(axis = 0)
        self.x_std = self.x.std(axis = 0, ddof = 1)

        # center
        if center == True:
            x_center = self.x - self.x_mean
        else:
            x_center = self.x
        
        # scale
        if scale == True:
            x_zscore = x_center/self.x_std
        else:
            x_zscore = x_center
        
        # data matrix
        data_mat = x_zscore

        # svd
        U, vals, V = np.linalg.svd(data_mat)
        vals = np.square(vals)
        vecs = V.T
        self.eig_vals = np.round(vals/np.sum(vals), 4)
        self.eig_vecs = vecs

        # 根据k，判定输出维度out_k
        if self.k >=1:
            self.out_k = int(self.k)
        elif self.k > 0 and self.k < 1:
            pc_explain = self.eig_vals.cumsum()
            cumratio_bool = pc_explain >= self.k
            tmp = np.where(cumratio_bool == True)[0][0] + 1
            self.out_k = tmp
 

    def fit_transform(self, x, center=True, scale=False):
        '''
        x: pca降维初始数据， np.arary数组，行表示样本，列表示特征
        '''
        self.x = np.asarray(x)
        self.x_mean = self.x.mean(axis = 0)
        self.x_std = self.x.std(axis = 0, ddof = 1)

        # center
        if center == True:
            x_center = x - self.x_mean
        else:
            x_center = x
        
        # scale
        if scale == True:
            x_zscore = x_center/self.x_std
        else:
            x_zscore = x_center
        
        # cov matrix
        data_mat = x_zscore

        # svd
        U, vals, V = np.linalg.svd(data_mat)
        vals = np.square(vals)
        vecs = V.T
        self.eig_vals = np.round(vals/np.sum(vals), 4)
        self.eig_vecs = vecs

        # 根据k，判定输出维度out_k
        if self.k >=1:
            self.out_k = int(self.k)
        elif self.k > 0 and self.k < 1:
            pc_explain = self.eig_vals.cumsum()
            cumratio_bool = pc_explain >= self.k
            tmp = np.where(cumratio_bool == True)[0][0] + 1
            self.out_k = tmp

        # 在新维度上的结果
        new_x = np.matmul(x_zscore, self.eig_vecs)
        self.new_x = new_x

        return(new_x[:, 0:self.out_k])


    def transform(self, y, center=True, scale=False):
        '''
        y: pca降维验证数据， np.arary数组，行表示样本，列表示特征
        '''
        y = np.asarray(y)

        # center
        if center == True:
            y_center = y - self.x_mean
        else:
            y_center = y
        
        # scale
        if scale == True:
            y_zscore = y_center/self.x_std
        else:
            y_zscore = y_center

        # 在新维度上的结果
        new_y = np.matmul(y_zscore, self.eig_vecs)
        self.new_y = new_y

        return(new_y[:, 0:self.out_k])


# obj: data grouping (default is 13 sets of data)
class sampleSplit:
    def __init__(self, arg):
        self.arg = arg
    def CVSplit(self, Xdf, Ydf):
        all_test = []
        all_split = []
        if self.arg.crossV == 1:
            all_split = [[range(Xdf.shape[0]), range(Xdf.shape[0])]]
        else:
            CVS = CrossvalidationSplit(
                n_splits=self.arg.crossV, test_size=self.arg.testS, CVt=self.arg.CVfit, n_repeats=2, leavP=self.arg.leavP, random_state=self.arg.random)
            SFA = CrossvalidationSplit(
                n_splits=self.arg.crossV, test_size=self.arg.testS, CVt='SFA', n_repeats=2, leavP=self.arg.leavP, random_state=self.arg.random)
            for train_index, test_index in CVS.split(Xdf, Ydf):
                all_test.extend(test_index)
                all_split.append([train_index.tolist(), test_index.tolist()])
            # if the unique of all test sample is not equal to input sample, add other 3 sets data by SFA method spliting
            if self.arg.CVfit == 'SSA' and not self.arg.fit_10:
                if len(set(all_test)) < len(Ydf):
                    all_add = [[tr.tolist(), te.tolist()]
                               for tr, te in SFA.split(Xdf, Ydf)]
                    all_split += all_add
        return(all_split)


class HyperparamOptimal():
    def __init__(self, arg, log, model, y_variable_type="C",  y_variable_class=2, score= None):
        self.arg = arg
        self.log = log
        self.model = model
        self.Type = y_variable_type
        self.MClass = y_variable_class
        self.score = score
    
    def Pdecorator(self):
        # sample split
        if self.Type == 'C':
            self.SSS = CrossvalidationSplit(n_splits=10, test_size=self.arg.GStestS, CVt='SSS', random_state=self.arg.random)
        elif self.Type == 'R':
            self.SSS = CrossvalidationSplit(n_splits=3, n_repeats=3, CVt='RSKF', random_state=self.arg.random)
        else:
            self.SSS = ''
        # get estimator and parameters
        self.estimator = ML(self.model, Type=self.Type).GetPara().estimator
        if self.arg.SearchCV:
            self.parameters = ML(self.model, Type=self.Type, SearchCV=self.arg.SearchCV).GetPara().parameters
        else:
            self.parameters = {}
        
    def Hyperparemet(self, x_train, x_test, y_train, y_test):
        HyperparamOptimal.Pdecorator(self)
        self.x_train = x_train
        self.x_test = x_test
        self.y_train = y_train
        self.y_test = y_test
        self.log += HyperparamOptimal.logRecord('info', 'hyperparameter optimization in the %s model......\n', self.model)
        if (self.MClass > 2) & (self.model in ['XGB']):
            self.estimator = self.estimator.set_params(objective='multi:softprob')
        if self.model in ['XGB']:
            _n_jobs = 20
            if 'scale_pos_weight' in self.parameters.keys():
                pos_weight = y_train.value_counts(normalize=True)
                self.parameters['scale_pos_weight'] += ((1-pos_weight)/pos_weight).to_list()
                self.parameters['scale_pos_weight'] = list(set(self.parameters['scale_pos_weight']))
        else:
            _n_jobs = self.arg.n_job
        # add hyperopt Spearmint hyperparameter
        if not self.arg.SearchCV:
            clf = self.estimator
            self.log += HyperparamOptimal.logRecord('info','decrepate hyperparameter optimization......\n')
        elif self.arg.SearchCV == 'GSCV':
            clf = GridSearchCV(
                estimator=self.estimator,
                param_grid=self.parameters,
                n_jobs=_n_jobs,
                cv=self.SSS,
                scoring=self.score,
                error_score=np.nan,
                return_train_score=True,
                refit=True
            )
            self.log += HyperparamOptimal.logRecord('info','GSCV hyperparameter optimization......\n')
        elif self.arg.SearchCV == 'RSCV':
            clf = RandomizedSearchCV(
                estimator=self.estimator,
                param_grid=self.parameters,
                n_jobs=_n_jobs,
                cv=self.SSS,
                n_iter=self.arg.n_iter,
                scoring=self.score,
                return_train_score=True,
                refit=True,
                error_score='raise'
            )
            self.log += HyperparamOptimal.logRecord('info','RSCV hyperparameter optimization......\n')
        # fit
        if self.model in ['XGB_']:
            clf.fit(x_train, y_train,
                    #eval_metric=["error", "logloss"],
                    #eval_set=[(X_train, y_train), (X_test, y_test)],
                    eval_set=[(x_test, y_test)],
                    eval_metric='auc',
                    early_stopping_rounds=15,
                    verbose=False)
        else:
            clf.fit(x_train, y_train)
        # print best parameters
        if hasattr(clf, 'best_estimator_'):
            self.log += HyperparamOptimal.logRecord('info','%s best parameters in %s: \n%s\n', self.model, self.arg.SearchCV, clf.best_estimator_.get_params())
            self.clf = clf.best_estimator_
        else:
            self.log += HyperparamOptimal.logRecord('info','%s best parameters: \n%s\n', self.model, clf.get_params())
            self.clf = clf

    # obj: refitting the models by the entire train data
    @staticmethod
    def re_fit(clf, x_matrix, y_matrix, model, log):
        new_clf = clf.fit(x_matrix, y_matrix)
        log += HyperparamOptimal.logRecord('info', '%s model has been refitted by input train data.\n', model)
        return new_clf
    
    @staticmethod
    def logRecord(*args):
        if args[0] == 'info':
            return '[%s] [INFO]: ' % time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()) + args[1] % args[2:]
        elif args[0] == 'warning':
            return '[%s] [WARNING]: ' % time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()) + args[1] % args[2:]

    @staticmethod
    def CalibratedClassifierCV(clf, CV, method, x_matrix, y_matrix):
        clf_C = CalibratedClassifierCV(clf, cv=CV, method=method)
        clf_C.fit(x_matrix, y_matrix)
        return clf_C

    @staticmethod
    def MergeTrainTest(train, test):
        try:
            allData = pd.concat((train, test), axis=0)
        except TypeError:
            allData = vstack((train, test))
        return allData

    @staticmethod
    def GetPredict(clf, x_matrix, mclass, log):
        predict = clf.predict(x_matrix)
        try:
            proba = clf.predict_proba(x_matrix)
        except AttributeError:
            proba = clf._predict_proba_lr(x_matrix)
            log += HyperparamOptimal.logRecord('warning','Note: LinearSVC, SGB use _predict_proba_lr based on decision_function as predict_proba in GridsearchCV.\n')
        except:
            proba = _predict_proba_lr(clf.decision_function(x_matrix))
            log += HyperparamOptimal.logRecord('warning','Note: predict_proba use sigmoid transversion based on decision_function!\n')

        if (mclass & mclass != proba.shape[1]):
            raise Exception('The columns length of predict probability is wrong!') 
        return np.c_[predict, proba], log

    @staticmethod
    def GetCoef(model, clf, mclass, log, index=None):
        if model in ['MNB','BNB'] :
            importances= np.exp(clf.coef_)
            log += HyperparamOptimal.logRecord('warning','Note: Use coef_ exp values as the feature importance of the estimator %s.\n', model)
        elif model in ['CNB'] :
            if mclass ==2:
                importances= np.exp(-clf.feature_log_prob_)[1]
            else:
                importances= np.exp(-clf.feature_log_prob_)
            log += HyperparamOptimal.logRecord('warning','Note: Use feature_log_prob_ negative exp values as the feature importance of the estimator %s.\n', model)
        elif model in ['GNB'] :
            if mclass ==2:
                importances= clf.theta_[1]
            else:
                importances= clf.theta_
            log += HyperparamOptimal.logRecord('warning','Note: Use theta_ values as the feature importance of the estimator %s.\n', model)
        elif model in ['MLP'] :
            def collapse(coefs):
                Coefs = coefs[0]
                for b in coefs[1:]:
                    Coefs = Coefs.dot(b)
                return(Coefs/Coefs.sum(0))
            importances = collapse(clf.coefs_).T
            log += HyperparamOptimal.logRecord('warning','Note: Use coefs_ weighted average as the feature importance of the estimator %s.\n', model)
        elif model in ['SVM', 'SVMrbf', 'nuSVMrbf']:
            dot_coef_ = clf.dual_coef_.dot( clf.support_vectors_ )
            importances = (1-np.exp(-dot_coef_)) / (1+np.exp(-dot_coef_))
            log += HyperparamOptimal.logRecord('warning','Note: Use exp dot of support_vectors_ and dual_coef_ values as the feature importance of the estimator %s.\n', model)
        else:
            for i in ['feature_importances_', 'coef_']:
                try:
                    importances= eval('clf.%s'%i)
                    log += HyperparamOptimal.logRecord('warning','Note: Use %s as the feature importance of the estimator %s.\n', i, model)
                    break
                except AttributeError:
                    importances = []
                    log += HyperparamOptimal.logRecord('warning','Note: Cannot find the feature importance attributes of estimator %s.\n', i)
        df_import = pd.DataFrame(np.array(importances).T)
        if not df_import.empty :
            df_import.index= index
            df_import = df_import[(df_import != 0).any(1)]
        log += HyperparamOptimal.logRecord('info', '%s Feature coefficiency: \n%s\n', model, df_import)
        return df_import, log

    @staticmethod
    def OneHot(model, clf, x_matrix, OHE='N'):
        if model in ['GBDT']:
            x_trans = clf.apply(x_matrix)[:,:,0]
        elif model in ['XGB', 'RF']:
            x_trans = clf.apply(x_matrix)
        if OHE == 'N':
            OHE = OneHotEncoder(categories='auto')
            OHE.fit(x_trans)
        x_trans = OHE.transform(x_trans.astype(np.int32))
        return [x_trans, OHE]

    @staticmethod            
    def GetEvalu_C(y_true, y_predict, y_score, name=None):
        y_trueb = Check_Binar(y_true)

        Model_ = pd.Series({  
                    'Accuracy'  : accuracy_score(y_true, y_predict),
                    'Accuracy_B': balanced_accuracy_score(y_true, y_predict),
                    'F1_score'  : f1_score(y_true, y_predict),
                    'Precision' : precision_score(y_true, y_predict),
                    'Recall'    : recall_score(y_true, y_predict),
        })

        Model_['Roc_auc'] = round(roc_auc_score(y_trueb, y_score), 6)
        Model_['Precision_A'] = round(average_precision_score( y_trueb, y_score), 6)
        if y_score.shape[1] > 2:
            for i in range(y_score.shape[1]):
                Model_['Roc_auc_' + str(i)] = round(roc_auc_score(y_trueb[:,i], y_score[:, i]), 6)
                Model_['Precision_A_' + str(i)] = round(average_precision_score( y_trueb[:, i], y_score[:, i]), 6)

        Model_.name = name
        return Model_.sort_index()
    
    @staticmethod
    def GetScore(df, label=None):
        _m = [re.sub('^Pred','', _i) for _i in df.columns if re.match('Pred', _i)]
        return pd.concat([HyperparamOptimal.GetEvalu_C(df['TRUE'],
                                                       df['Pred'+_j],
                                                       df[sorted([_n for _n in df.columns if re.search('Prob'+_j, _n)])],
                                                       name= label + _j) for _j in _m], axis=1)
                                                 
    def getModelResult(self, x_matrix='N'):
        if x_matrix=='N':
            self.x = HyperparamOptimal.MergeTrainTest(train=self.x_train, test=self.x_test)
        else:
            self.x = x_matrix
        self.y = HyperparamOptimal.MergeTrainTest(train=self.y_train, test=self.y_test)
        # self.new_clf = HyperparamOptimal.re_fit(clf=self.clf, x_matrix=self.x, y_matrix=self.y, model=self.model, log=self.log)
        self.clf_C = HyperparamOptimal.CalibratedClassifierCV(clf=self.clf, CV=self.SSS, method=self.arg.calibme, x_matrix=self.x_train, y_matrix=self.y_train)
        predict_G, self.log = HyperparamOptimal.GetPredict(clf=self.clf, x_matrix=self.x, mclass=self.MClass, log=self.log)
        predict_C, self.log = HyperparamOptimal.GetPredict(clf=self.clf_C, x_matrix=self.x, mclass=self.MClass, log=self.log)
        predict = pd.DataFrame(np.c_[self.y, predict_G, predict_C], index=self.y.index)
        predict.columns = ['TRUE', 'Pred_%s'%self.model] + ['%s_Prob_%s'%(x, self.model) for x in range(self.MClass)] + \
                            ['Pred_C_%s'%self.model] + ['%s_Prob_C_%s'%(x, self.model) for x in range(self.MClass)]
        coef, self.log = HyperparamOptimal.GetCoef(model=self.model, clf=self.clf, mclass=self.MClass, log=self.log, index=self.x.columns)
        Best_MD = [{'model':self.model, 'clf':self.clf, 'features':self.x.columns.tolist()},
                   {'model':'C_'+self.model, 'clf':self.clf_C, 'features':self.x.columns.tolist()}]
        return(predict, coef, Best_MD)

    # obj: predict the testing dataset by 13 models
    def Predicting_C(self, best_model, pDFall, y_name, xa):
        Y_TRUE  = pDFall[y_name].to_frame(name='TRUE')
        Y_PRED , Y_Eval = [], []
        for _k, _cv in enumerate(best_model):
            y_pred = [Y_TRUE]
            for _i, _m in enumerate(_cv):
                model = _m['model']
                clf = _m['clf']
                model_x_names = _m['features']
                if 'pca_clf' in _m:
                    pca_clf = _m['pca_clf']
                    tmp = pd.DataFrame(pca_clf.transform(pDFall[xa], center=self.arg.pc_center, scale=self.arg.pc_scale), index=pDFall.index)
                    tmp.columns = ["PC"+str(j) for j in range(1, tmp.shape[1]+1)]
                    x_pred = tmp[model_x_names]
                else:
                    x_pred = pDFall[model_x_names]
                if 'leaf' in _m.keys():
                    leaf = _m['leaf']
                    x_pred, OHE = HyperparamOptimal.OneHot(self.model, _cv[0]['clf'], x_pred, leaf)
                log = ''
                _prepro, log = HyperparamOptimal.GetPredict(clf, x_pred, self.MClass, log)
                self.log.NIF(log)
                _prepro = pd.DataFrame(_prepro,
                                       index=pDFall.index,
                                       columns = ['Pred_%s' % model] + ['%s_Prob_%s'%(x, model) for x in range(self.MClass)]
                                       )
                y_pred.append(_prepro)
            y_pred  = pd.concat(y_pred, axis=1)
            Openf('%s%s_Class_%s_predict.xls'%(self.arg.output, y_name, "Test_"+str(_k+1)), (y_pred), index=True, index_label='sample').openv()
            y_predv = y_pred[~y_pred['TRUE'].isna()]
            self.log.CIF(('%s %s Model Predicting Parameters'% (y_name, self.model)).center(45, '-'))
            if y_predv['TRUE'].nunique() >= 2:
                Pred_Mel = HyperparamOptimal.GetScore(df=y_predv, label='Predict')
                self.log.CIF("%s Modeling evaluation: \n%s" % (self.model, Pred_Mel))
            else:
                Pred_Mel = pd.DataFrame()
            self.log.CIF(('Completed %2d%%'% ((_k+1)*100/len(best_model))).center(45, '-'))
            Y_PRED.append(y_pred)
            Y_Eval.append(Pred_Mel)
        return(Y_PRED, Y_Eval)


class MergeAllResult():
    '''
    combined and integrated predicted values of 13 models.
    '''
    def __init__(self, log):
        self.log = log
        
    @staticmethod
    def revisedPred(All_Pb, All_Pred):
        _m = All_Pb.columns.str.replace('.*_Prob','Prob',regex=True).drop_duplicates(keep='first')
        for _x in _m:
            All_Pred[_x.replace('Prob','Pred')] = All_Pb.filter(regex=_x, axis=1).values.argmax(1)
        return All_Pred
    
    @staticmethod
    def calculate_MAD(predict):
        '''
        calculate MAD for each sample 13 models' predict value, the central point is median of 13 predict values.
        '''
        predict =list(predict)
        predict.sort()
        half_len = len(predict) // 2
        predict_median = (predict[half_len] + predict[~half_len])/2
        num = 0
        for i in predict:
            num += abs(i - predict_median)
        return num / len(predict)
    
    @staticmethod
    def mad_df(df):
        '''
        calculate MAD for dataframe by self-defining function calculate_MAD
        '''
        return df.apply(MergeAllResult.calculate_MAD)

    def CVMerge(self, predict, label='Test'):
        allTest = pd.concat(predict, axis=0)
        _model = [re.sub('^Pred_','', i) for i in allTest.columns if re.match('Pred_.*', i)]
        All_Prob_mean = allTest.filter(regex=("_Prob_")).groupby([allTest.index]).mean()
        All_Prob_mean.columns +=  '_mean'
        All_Prob_median = allTest.filter(regex=("_Prob_")).groupby([allTest.index]).median()
        All_Prob_median.columns += '_median'
        All_Prob_mad = allTest.filter(regex=("_Prob_")).groupby([allTest.index]).apply(MergeAllResult.mad_df)
        All_Prob_mad.columns += '_mad'
        All_Pred =  allTest.filter(regex=("^TRUE|^Pred_")).groupby([allTest.index]).apply(lambda x : x.mode(0).loc[0,:])
        All_Pred.columns = [ i + '_mode' if i !='TRUE' else i for i in All_Pred.columns ]
        All_Pred = MergeAllResult.revisedPred(All_Prob_mean, All_Pred)
        All_Pred = MergeAllResult.revisedPred(All_Prob_median, All_Pred)
        ALL_Result  = pd.concat([All_Pred, All_Prob_mean, All_Prob_median, All_Prob_mad], axis=1)
        ALL_Resultv = ALL_Result[~ ALL_Result['TRUE'].isna()]
        if ALL_Resultv.shape[0] > 2:
            EvalueA = HyperparamOptimal.GetScore(ALL_Resultv.filter(regex='^TRUE|_mean$', axis=1), label=label)
            self.log.CIF("The final mean evaluation: \n%s" % EvalueA)
        else:
            EvalueA = pd.DataFrame([])
        self.log.CIF(('Finish %s'% label).center(45, '-'))
        self.allTest = allTest
        self.ALL_Result = ALL_Result
        self.EvalueA = EvalueA
    
    def FeatCoeffs(self, All_import):
        if All_import:
            All_import = pd.concat(All_import, axis=1, sort=False).fillna(0)
            column = sorted(set(All_import.columns))
            All_import = All_import[column]
            for i in column:
                All_import['%s_mean'%i]  = All_import[i].mean(axis=1)
                All_import['%s_std'%i]   = All_import[i].std(axis=1)
                All_import['%s_median'%i]= All_import[i].median(axis=1)
            All_import.sort_values(by=['%s_mean'%i for i in column], ascending=[False]*len(column), inplace=True, axis=0)
            self.All_import = All_import
        else:
            self.All_import = pd.DataFrame([])
    
    def ModelScore(self, All_evluat):
        if All_evluat:
            All_pvalues = pd.concat(All_evluat,  axis=0)
            self.All_pvalues = All_pvalues
    
 
### define some functions
def modeling(arg, x_train, x_test, y_train, y_test, Typi, mclass, _k, _n, allGroup):
    log = ''
    hyper = HyperparamOptimal(arg, log, model=arg.model, y_variable_type=Typi, y_variable_class=mclass)
    hyper.Hyperparemet(x_train, x_test, y_train, y_test)
    predict, coef, Best_MD = hyper.getModelResult()
    logText = hyper.log
    # addmodel
    if (arg.model in ['GBDT', 'XGB', 'RF']) and (arg.Addmode !='N'):
        log_add = ''
        x_train_add, OHE = HyperparamOptimal.OneHot(model=arg.model, clf=hyper.clf, x_matrix=x_train)
        x_test_add,  OHE = HyperparamOptimal.OneHot(model=arg.model, clf=hyper.clf, x_matrix=x_test, OHE=OHE)
        hyper_add = HyperparamOptimal(arg, log_add, model=arg.Addmode, y_variable_type=Typi, y_variable_class=mclass)
        hyper_add.Hyperparemet(x_train_add, x_test_add, y_train, y_test)
        predict_add, coef_add, Best_MD_add = hyper_add.getModelResult()
        predict = pd.concat([predict, predict_add], axis=1)
        Best_MD_add[0].update({'leaf':OHE})
        Best_MD_add[1].update({'leaf':OHE})
        Best_MD += Best_MD_add
    # statistic    
    if y_test.nunique() >= 2 and mclass<=2:
        score_train = HyperparamOptimal.GetScore(predict.loc[y_train.index], label='Train')
        score_validation = HyperparamOptimal.GetScore(predict.loc[y_test.index], label='Validation')
        score = pd.concat([score_train, score_validation], axis=1)
    else:
        score = pd.DataFrame()
    logText += HyperparamOptimal.logRecord('info', '%s Modeling evaluation: \n%s\n', arg.model, score)
    logText += HyperparamOptimal.logRecord('info', ('Modeling has been Completed %2d%%%%' % ((_k+1)*(_n+1)*100/(arg.Repeatime*allGroup))).center(45, '-'))
    logText += "\n"
    return(predict, coef, score, Best_MD, logText)

def run_fitting(args, Log):
    ### set fitting output directory
    args.output = '%s/01ModelFit/%s' % (args.outdir, args.header)
    os.makedirs(os.path.dirname(args.output), exist_ok=True)
    os.makedirs(os.path.dirname(args.MODELF), exist_ok=True)

    ### load data
    (group, RYa, CYa, Xa, Xg) = OpenM(args, Log).openg()
    YType = group[(group.Variables.isin(RYa + CYa))][['Variables', 'Type']]
    for Yi, Typi in YType.values:
        DFall = Openf(args.input, index_col=0).openb()
        Xi    = list(set(DFall.columns) - set([Yi]))
        Xi.sort(key=DFall.columns.to_list().index)
        Log.CIF(('%s: Supervised MODELing' % Yi).center(45, '*') )
        Log.NIF('%s Value Counts:\n%s' % (Yi, DFall[Yi].value_counts().to_string())) 
        DFall[Yi] = Check_Label(DFall[Yi])
        # sample was splited into 13 groups
        sample_split = sampleSplit(args)
        sample_group_index = sample_split.CVSplit(Xdf=DFall[Xi], Ydf=DFall[Yi])
        # hyperparameter optimization (use the first set of data)
        allPredict, allCoef, allScore, allMD = [], [], [], []
        # threads = []
        for repeat in range(args.Repeatime):
            for i in range(len(sample_group_index)):
                train_index, test_index = sample_group_index[i]
                train , test = DFall.iloc[train_index, :], DFall.iloc[test_index, :]
                x_train, y_train = train[Xi], train[Yi]
                x_test , y_test  = test[Xi], test[Yi]
                if args.pc_number != 0:
                    if args.pc_number >= 1:
                        pc_number = int(args.pc_number)
                    else:
                        pc_number = args.pc_number
                    print(pc_number)
                    pca_clf = pca_svd(k=pc_number)
                    pca_clf.fit(x_train, center=args.pc_center, scale=args.pc_scale)
                    print(pca_clf.out_k)
                    x_train = pd.DataFrame(pca_clf.transform(x_train, center=args.pc_center, scale=args.pc_scale), index=x_train.index)
                    x_test = pd.DataFrame(pca_clf.transform(x_test, center=args.pc_center, scale=args.pc_scale), index=x_test.index)
                    x_train.columns = ["PC"+str(j) for j in range(1, x_train.shape[1]+1)]
                    x_test.columns = ["PC"+str(j) for j in range(1, x_test.shape[1]+1)]
            #    t = MyThread(modeling, (args, x_train, x_test, y_train, y_test, Typi, DFall[Yi].nunique(), repeat, i, len(sample_group_index)))
            #    threads.append(t)
            #for i in range(len(sample_group_index)):
            #    threads[i].setDaemon(True)
            #    threads[i].start()
            #for i in range(len(sample_group_index)):
            #    threads[i].join()
            #for i in range(len(sample_group_index)):
            #    full_predict, coef, score, MD, logText = threads[i].get_result()
                full_predict, coef, score, MD, logText = modeling(args, x_train, x_test, y_train, y_test, Typi, DFall[Yi].nunique(), repeat, i, len(sample_group_index))
                Openf('%s%s_Class_%s_predict.xls'%(args.output, Yi, "Train_"+str(i+1)), (full_predict.loc[y_train.index, ]), index=True, index_label='sample').openv()
                Openf('%s%s_Class_%s_predict.xls'%(args.output, Yi, "Validation_"+str(i+1)), (full_predict.loc[y_test.index, ]), index=True, index_label='sample').openv()
                allPredict.append(full_predict)
                if args.pc_number != 0:
                    for j in range(0, len(MD)):
                        MD[j].update({'pca_clf':pca_clf})
                allMD.append(MD)
                if not coef.empty:
                    allCoef.append(coef)
                if not score.empty:
                    allScore.append(score)
                Log.NIF(logText)
            Log.CIF('Modeling has been Completed'.center(45, '-'))
        Log.CIF(('%s: Supervised MODELing Finish'%Yi).center(45, '*'))
        # merge all test predict
        matp = MergeAllResult(Log)
        matp.CVMerge(allPredict, label='Train')
        matp.FeatCoeffs(allCoef)
        matp.ModelScore(allScore)
        Openf('%s%s_Class_%s_detials.xls'%(args.output, Yi, 'Train'), (matp.allTest), index=True, index_label='sample').openv()
        Openf('%s%s_Class_%s_summary.xls'%(args.output, Yi, 'Train'), (matp.ALL_Result), index=True, index_label='sample').openv()
        Openf('%s%s_Class_%s_summary_statistics.xls'%(args.output, Yi, 'Train'), (matp.EvalueA), index=True, index_label='Score').openv()
        Openf('%s%s_Class_features_importance.xls' %(args.output, Yi), (matp.All_import)).openv()
        Openf('%s%s_Class_%s_metrics_pvalues.xls'%(args.output, Yi, 'Train'), (matp.All_pvalues), index=True, index_label='Score').openv()
        joblib.dump(DFall[Xi+[Yi]], '%s%s_Class_traintest_set.pkl' %(args.MODELF, Yi), compress=1)
        joblib.dump(allMD, '%s%s_Class_best_estimator_1.pkl' %(args.MODELF, Yi), compress=1)
    Log.CIF(('%s: Supervised MODELing Finish' % Yi).center(45, '*'))
   
if __name__ == "__main__":
    from MLLogging import Logger
    from MLArguments import Args
    args = Args()
    args.commands = "Fitting"
    args.input = "../Data/fragment.stat_Methy.Ratio.PCA_0.95.discovery.PCA.txt"
    args.group = "../Data/fragment.stat_Methy.Ratio.PCA_0.95.group.Info.PCA.txt"
    args.outdir = "../Test"
    args.model = "LinearSVM"
    args.Addmode = "N"
    args.Repeatime = 1
    args.CVfit = 'SSA'
    args.crossV = 10
    args.testS = 0.3
    args.GStestS = 0.3
    args.random = 123456
    args.leavP = 1
    args.calibme = 'sigmoid'
    args.n_job = 5
    args.SearchCV = 'GSCV'
    args.header = '/%s_' % (args.model)
    args.outdir = '%s/%s/' % (args.outdir, args.model)
    args.output = '%s%s'%(args.outdir, args.header)
    args.MODELF = '%s/00MODEL/%s' % (args.outdir, args.header)
    os.makedirs(os.path.dirname(args.outdir), exist_ok=True)
    os.makedirs(os.path.dirname(args.MODELF), exist_ok=True)
    # print output message and create new folder for modeling results
    Log = Logger( '%s%s_log.log'%(args.output, args.commands) )
    Log.NIF("The argument you have set as follows:".center(59, '*'))
    for i,k in enumerate(vars(args),start=1):
        Log.NIF('**%s|%-13s: %s'%(str(i).zfill(2), k, str(getattr(args, k))) )
    Log.NIF(59 * '*')

    run_fitting(args, Log)


from sklearn.linear_model import Lasso, LassoCV, ElasticNetCV, LassoLarsIC, SGDClassifier, LogisticRegressionCV,LogisticRegression
from sklearn.neighbors import KNeighborsClassifier,RadiusNeighborsClassifier
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC,SVR,LinearSVC,NuSVC,NuSVR
from sklearn.tree import DecisionTreeClassifier,DecisionTreeRegressor
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier, RandomForestRegressor,GradientBoostingClassifier,GradientBoostingRegressor
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis as QDA
from sklearn.naive_bayes import BernoulliNB, GaussianNB, MultinomialNB, ComplementNB
from scipy.stats import randint,uniform
from xgboost import XGBClassifier, XGBRegressor
#from catboost import CatBoostClassifier, CatBoostRegressor
from lightgbm import LGBMClassifier, LGBMRegressor
from skrvm import RVR, RVC
import numpy as np
#from bartpy.sklearnmodel import SklearnModel
#from bartpy.extensions.baseestimator import ResidualBART

class ML():
    def __init__(self, modle, Type='C', SearchCV='GSCV', *array, **dicto):
        self.modle = modle
        self.Type  = Type
        self.SearchCV = SearchCV
        self.array = array
        self.dicto = dicto

    def GetPara(self):
        if self.Type == 'C' :
            Parameters = self.Classification()
        elif self.Type == 'R' :
            Parameters = self.Regression()

        self.estimator  =  Parameters[self.modle]['estimator']
        self.parameters =  Parameters[self.modle]['parameters'][self.SearchCV]
        return self

    def Classification(self):
        return {
        'RF'   : {
            'estimator' : RandomForestClassifier(oob_score=True, n_estimators=100, n_jobs=10, random_state=123456),
            'parameters': {
                'GSCV'  : { 'max_features'     : [0.6, 0.7, 0.8, 0.9, 'sqrt','log2', None],
                            'max_depth'        : [3, 4, None],
                            'class_weight'     : ['balanced', 'balanced_subsample', None], 
                            'min_samples_split': [2, 4, 6 ],
                            'min_samples_leaf' : [1, 2, 3 ],
                            'criterion'        : ['gini', 'entropy'] },
                'RSCV'  : { 'max_features': [ 0.6, 0.7, 0.8, 0.9, 'sqrt','log2', None ],
                            'max_depth'   : [ 3, 4, None ],
                            'criterion'   : ['gini', 'entropy'],
                            'class_weight': ['balanced', 'balanced_subsample', None], 
                            'min_samples_split' : [ 2, 4, 6, 8, 11, 15],
                            'min_samples_leaf'  : [ 1, 2, 3, 7, 11, 15],
                            }
                }},

        'GBDT'          : {
            'estimator' : GradientBoostingClassifier(n_estimators=100, random_state=123456),
            'parameters': {
                'GSCV'  : { 'max_features': (0.5, 0.75, 0.8, 0.9, 'auto'),
                            'loss' : ['deviance', 'exponential'],
                            'max_depth': [ 3, 4, 5 ],
                            'learning_rate': [0.005, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9],
                            'min_samples_split':[2, 3, 4, 5],
                            'subsample': [0.75, 0.85, 0.95 ]},
                'RSCV'  : { 'max_features': (0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 'auto', 'log2', None),
                            'loss' : ['deviance', 'exponential'],
                            'max_depth': [ 3, 4, 5, 6, 7, 8, 9, 11, 15],
                            'learning_rate': [0.005, 0.01, 0.1, 0.2, 0.3, 0.5, 0.7, 0.9],
                            'min_samples_split':[2, 3, 4, 5, 6, 8],
                            'subsample': [0.7, 0.8, 0.9, 1]},
                }},

        'XGB'           : {
            'estimator' : XGBClassifier(
                            objective    ='binary:logistic', # multi:softprob
                            booster      ='gbtree',
                            silent       =True,
                            missing      =None,
                            n_jobs       =8,
                            random_state =123456),
            'parameters': {
                'GSCV'  : { 'colsample_bytree' : [0.75, 0.85, 0.95 ],
                            'subsample'        : [0.75, 0.85, 0.95 ],
                            'reg_alpha'        : [0, 0.1, 0.5, 1, 2  ],
                            'reg_lambda'       : [0.1, 0.5, 1, 2, 2.5 ],
                            'max_depth'        : [3, 4, 5, 6],
                            'learning_rate'    : [0.01, 0.05, 0.1, 0.2],
                            #'min_child_weight' : [1, 2],
                          },

                'RSCV'  : { 'subsample'        : [0.7, 0.8, 0.85],
                            'colsample_bytree' : [0.7, 0.8, 0.9],
                            'colsample_bylevel': [0.7, 0.8, 0.9],
                            'max_delta_step'   : [0, 1],
                            'colsample_bynode' : [1],
                            'scale_pos_weight' : [0.8, 1, 1.2],
                            'base_score'       : [0.5],
                            'gamma'            : [0, 0.005, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9],
                            'min_child_weight' : [1, 2, 3, 4, 5],
                            'min_samples_split': [2, 3, 4, 5, 6, 7],
                            'max_depth'        : [3, 4, 5, 6, 7, 8],
                            'reg_lambda'       : [0.2, 0.3, 0.5, 0.7, 0.8, 0.9, 0.96, 1],
                            'reg_alpha'        : [0, 0.001, 0.01, 0.1, 0.2, 0.3, 0.5, 0.7],
                            'learning_rate'    : [0.005, 0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5],
                           },
                }},

        'LGBM'          : {
            'estimator' : LGBMClassifier(
                            boosting_type='gbdt',
                            n_estimators=100,
                            min_split_gain=0.0,
                            min_child_weight=1e-3,
                            min_child_samples=10,
                            subsample_freq=0,
                            random_state=123456,
                            n_jobs=-1,
                            silent=False,
                            importance_type='split',
                            verbose = -1,
                            ),
            'parameters': {
                'GSCV'  : { 'num_leaves'       : [9, 17, 33, 65],
                            'max_depth'        : [-1, 3, 4, 5, 6],
                            'learning_rate'    : [0.01, 0.05, 0.1, 0.2],
                            'reg_alpha'        : [0.01, 0.1, 0.5, 1, 1.5, 2],
                            'reg_lambda'       : [0.01, 0.1, 0.5, 1, 1.5, 2],
                            'class_weight'     : ['balanced', None, ],
                            'subsample'        : [0.7, 0.78, 0.85],
                            'colsample_bytree' : [0.7, 0.78, 0.85],
                          },
                'RSCV'  : { 'num_leaves'       : [9, 17, 33, 65, 129],
                            'max_depth'        : [-1, 3, 4, 5, 6, 7],
                            'learning_rate'    : [0.01, 0.05, 0.1, 0.2],
                            'reg_alpha'        : [0, 0.01, 0.1, 0.5, 0.75, 1, 1.5, 2, 2.5, 3],
                            'reg_lambda'       : [0, 0.01, 0.1, 0.5, 0.75, 1, 1.5, 2, 2.5, 3],
                            'class_weight'     : ['balanced', None, ],
                            'subsample'        : [0.7, 0.78, 0.85],
                            'colsample_bytree' : [0.7, 0.78, 0.85],
                          }
                }},

        'AdaB_DT'   : {
            'estimator' : AdaBoostClassifier(DecisionTreeClassifier(criterion='gini', splitter='best',
                            class_weight='balanced', min_samples_leaf =1, max_features=None), random_state=123456),
            'parameters': {
                'GSCV'  : { 'n_estimators': range(400, 800, 1000),
                            'algorithm': ['SAMME', 'SAMME.R'],
                            "learning_rate": [0.3, 0.5, 0.7, 0.9, 0.95, 1, 2]},
                'RSCV'  : {}
                }},

        'MLP'           : {
            'estimator' : MLPClassifier(max_iter=15000,
                                        tol =1e-4,
                                        random_state=123456),
            'parameters': {
                'GSCV'  : { 'alpha': [10**i for i in [-6, -5, -4, -3, -1, 1]],
                            'solver':['adam', 'lbfgs', 'sgd'],
                            'activation': ["logistic", 'identity', "relu", "tanh"],
                            'hidden_layer_sizes': [ (50, 30, 20, 10, 10,  ), (15, 25, 15, 10, ), (8, 15, 10, 10, ), (30, 10, 40, 10, ), (10, 20, 10, 10, )],
                            'learning_rate': ["constant", "invscaling", "adaptive"]},
                'RSCV'  : {'hidden_layer_sizes': [ (randint.rvs(8,30,1), randint.rvs(5,20,1), randint.rvs(5,30,1),) ,
                                                   (randint.rvs(8,40,1), randint.rvs(5,40,1), )],
                            'activation': ["logistic", 'identity', "relu", "tanh"],
                            'solver':['adam', 'lbfgs', 'sgd'],
                            'alpha': [10**i for i in [-6, -5, -4, -3, -1, 1]], #uniform(1e-06, 0.9),
                            'learning_rate': ["constant", "invscaling", "adaptive"]}
                }},

        'LinearSVM'  : {
            'estimator' : LinearSVC(penalty='l2', dual=True, random_state=123456),
            'parameters': {
                'GSCV'  : { 'loss': ['hinge', 'squared_hinge'],
                            'tol': [1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2 ],
                            'C': np.power(10, np.linspace(-1.7, 1.7, 20)),
                            'max_iter':[int(8e6), int(1.5e7)]},
                'RSCV'  : {}
                }},
        'LinearSVMl1'   : {
            'estimator' : LinearSVC(penalty='l1', max_iter=int(1e6), random_state=123456),
            'parameters': {
                'GSCV'  : { 'dual' : [False, True],
                            'loss': ['squared_hinge'],
                            'tol': [ 1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2 ],
                            'C': np.power(10, np.linspace(-2, 2, 10))},

                'RSCV'  : {}
                }},
        'SVMlinear'     : {
            'estimator' : SVC(kernel="linear", probability=True, decision_function_shape='ovr', random_state=123456),
            'parameters': {
                'GSCV'  : [{'tol': [1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1 ],
                            'C': np.power(10, np.linspace(-2, 2, 10)),
                           }],
                'RSCV'  : {}
                }},
        'SVMrbf'        : {
            'estimator' : SVC(kernel='rbf',  probability=True, decision_function_shape='ovr', random_state=123456),
            'parameters': {
                'GSCV'  : { 'gamma': [1e0, 1e-1, 1e-2, 1e-3, 'scale'],
                            'tol': [ 1e-4, 1e-3, 1e-2 ],
                            'C': np.power(10, np.linspace(-2, 0, 10)),
                          },
                'RSCV'  : {}
                }},
        'SVMpoly'        : {
            'estimator' : SVC(kernel='poly',  probability=True, decision_function_shape='ovr', random_state=123456),
            'parameters': {
                'GSCV'  : { 'gamma': [1e0, 1e-1, 1e-2, 1e-3, 'scale'],
                            'tol': [ 1e-4, 1e-3, 1e-2 ],
                            'C': np.power(10, np.linspace(-2, 0, 10)),
                          },
                'RSCV'  : {}
                }},
        'RVMrbf'        : {
            'estimator' : RVC(kernel='rbf', n_iter=1000, threshold_alpha=1e9),
            'parameters': {
                'GSCV'  : { 'tol': [ 1e-5, 1e-4, 1e-3, 1e-2, 1e-1 ],
                            'alpha': [ 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6],
                            'beta': [ 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6],
                            'beta_fixed': [False],
                            'bias_used': [True],
                          },
                'RSCV'  : {}
                }},
        'RVMpoly'        : {
            'estimator' : RVC(kernel='poly', n_iter=1000, threshold_alpha=1e9),
            'parameters': {
                'GSCV'  : { 'tol': [ 1e-5, 1e-4, 1e-3, 1e-2, 1e-1 ],
                            'alpha': [ 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6],
                            'beta': [ 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6],
                            'beta_fixed': [False],
                            'bias_used': [True]
                          },
                'RSCV'  : {}
                }},
        'SVM'           : {
            'estimator' : SVC(probability=True, decision_function_shape='ovr', random_state=123456),
            'parameters': {
                'GSCV'  : [{'kernel': ['rbf', 'sigmoid'],
                            'gamma': [1e1, 1e0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 'scale'],
                            'tol': [1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1 ],
                            'C': np.power(10, np.linspace(-2, 2, 10)),
                           },
                           {'kernel': ['poly'],
                            'degree' : [2, 3, 4, 5],
                            'gamma': [1e1, 1e0, 1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 'auto'],
                            'tol': [1e-7, 1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1 ],
                            'C': np.power(10, np.linspace(-2, 2, 10)),
                           },
                          ],
                'RSCV'  : {}
                }},
        'nuSVMrbf'      : {
            'estimator' : NuSVC(probability=True, decision_function_shape='ovr', random_state=123456),
            'parameters': {
                'GSCV'  : { 'kernel': ['rbf'],
                            'nu': [0.8, 0.9, 1],
                            'gamma': [1e-2,1e-3,1e-4,1e-5,5e-2,5e-3,5e-4,'auto'],
                            'tol': [1e-6, 1e-5, 5e-5, 1e-4, 5e-3, 3e-3, 1e-3, 1e-2, 5e-2]},
                'RSCV'  : {}
                }},

        'SGD'           : {
            'estimator' : SGDClassifier(max_iter = int(1e5), learning_rate='optimal', n_jobs=2, random_state=123456),
            'parameters': {
                'GSCV'  : [{ 'penalty' : ['l2','elasticnet'],
                             'loss': ['hinge','log','modified_huber','squared_hinge','perceptron' ],
                             'alpha': [5e-5, 1e-4, 5e-4, 1e-3, 1e-2, 1e-1, 1e0, 1e1 ],
                             'l1_ratio': [0.01, 0.05, 0.1, 0.2, 0.3, 0.5, 0.7, 0.9 ],
                             #'eta0' : [0, 0.0001, 0.001, 0.01, 0.1],
                             'tol': [1e-3, 5e-3, 1e-4, 1e-5, 1e-2]},
                          ],
                'RSCV'  : {}
                }},

        'KNN'           : {
            'estimator' : KNeighborsClassifier(n_jobs=2),
            'parameters': {
                'GSCV'  : { 'n_neighbors':[3, 4, 5, 6, 7, 8, 10, 12, 15],
                            'weights': ['uniform', 'distance'],
                            'algorithm': ['auto','ball_tree','kd_tree','brute'],
                            'leaf_size': [10,20,30,40, 50, 60, 70, 100],
                            'p': [1,2]},
                'RSCV'  : {}
                }},

        'RNN'           : {
            'estimator' : RadiusNeighborsClassifier(metric='minkowski', outlier_label=None, n_jobs=2),
            'parameters': {
                'GSCV'  : { 'radius': [1,2,3,4,5,10,15,20,23,26,30,35,40],
                            'weights': ['uniform', 'distance'],
                            'algorithm':['auto','ball_tree','kd_tree','brute'],
                            'leaf_size': [10,20,30,50,70,100,200],
                            'p': [1,2]},
                'RSCV'  : {}
                }},

        'GNB'           : {
            'estimator' : GaussianNB(priors=None, var_smoothing=1e-09),
            'parameters': {
                'GSCV'  : {'var_smoothing' : np.dot( np.array([[1e-11, 1e-10,1e-09,1e-08,1e-07]]).T, np.array([[1,3,5,7]]) ).flatten() },
                'RSCV'  : {}
                }},

        'BNB'           : {
            'estimator' : BernoulliNB(fit_prior=True, class_prior=None),
            'parameters': {
                'GSCV'  : {'alpha':[i/10 for i in range(1,21,1)], 'binarize' : [i/10 for i in range(1,10,1)] },
                'RSCV'  : {}
                }},

        'MNB'           : {
            'estimator' : MultinomialNB(fit_prior=True, class_prior=None),
            'parameters': {
                'GSCV'  : {'alpha':[i/10 for i in range(1,21,1)]},
                'RSCV'  : {}
                }},

        'CNB'           : {
            'estimator' : ComplementNB(fit_prior=True, class_prior=None, norm=False),
            'parameters': {
                'GSCV'  : {'alpha':[i/10 for i in range(1,21,1)]},
                'RSCV'  : {}
                }},

        'DT'    : {
            'estimator' : DecisionTreeClassifier(splitter='best', class_weight='balanced', random_state=123456),
            'parameters': {
                'GSCV'  : { 'max_features': (0.4, 0.5, 0.6, 0.7, 0.8, 'sqrt','log2','None'),
                            'min_samples_leaf' : (1,2,3) ,
                            'max_depth': (5, 8, 10, 15, 25, 30, None),
                            'criterion' : ['gini', 'entropy']},
                'RSCV'  : {}
                }},

        'LR'            : {
            'estimator' : LogisticRegression(random_state=123456,
                                             fit_intercept=True,
                                             max_iter=10000,
                                             multi_class='auto'),
            'parameters': {
                'GSCV'  : [ { 'penalty'  : ['l1'],
                              'tol'      : [ 1e-3, 1e-4,  1e-5, ],
                              'l1_ratio' : [None],
                              'solver'   : ['liblinear', 'saga']},
                            { 'penalty'  : ['l2'],
                              'tol'      : [1e-3, 1e-4,  1e-5, ],
                              'l1_ratio' : [None],
                              'solver'   : ['lbfgs', 'sag']},
                            { 'penalty'  : ['elasticnet'],
                              'tol'      : [ 1e-3, 1e-4, 1e-5, ],
                              'l1_ratio' : [0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95],
                              'solver'   : ['saga']},
                          ],
                'RSCV'  : {}
                }},

        'LRCV'          : {
            'estimator' : LogisticRegressionCV(random_state=123456,
                                             fit_intercept=True,
                                             n_jobs = 30,
                                             multi_class='auto'),
            'parameters': {
                'GSCV'  : [ { 'penalty'  : [ 'elasticnet'],
                              'Cs'       : [ 10,
                                             np.power(10, np.arange(-4,4,0.4)),
                                             np.power(10, np.arange(-3,3,0.3))
                                           ],
                              'l1_ratios': [ [0.005, 0.05, 0.1, 0.3, 0.5, 0.7, 0.9], ],
                              'max_iter' : [ int(4e4) ],
                              'cv'       : [ 10  ],
                              'solver'   : [ 'saga']},
                          ],
                'RSCV'  : [ { 'penalty'  : [ 'l1'],
                              'Cs'       : [ np.power(10, np.arange(-2,4,0.3)) ],
                              'l1_ratios': [ None ],
                              'max_iter' : [ int(6e4) ],
                              'cv'       : [ 10 ],
                              'solver'   : [ 'liblinear', 'saga']},
                            { 'penalty'  : [ 'l2'],
                              'l1_ratios': [ None ],
                              'max_iter' : [ int(6e4) ],
                              'cv'       : [ 10   ],
                              'Cs'       : [ np.power(10, np.arange(-2,4,0.3)) ],
                              'solver'   : [ 'lbfgs', 'sag']},
                            { 'penalty'  : [ 'elasticnet'],
                              'Cs'       : [ np.power(10, np.arange(-2,4,0.3)) ],
                              'max_iter' : [ int(6e4) ],
                              'cv'       : [ 10 ],
                              'solver'   : ['saga']},
                          ],
                }},

        'LassoCV'       : {
            'estimator' : LassoCV(cv=10,
                                  alphas=[0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 1, 10],
                                  max_iter=20000,
                                  n_jobs=-1),
            'parameters': {
                'GSCV'  : { 'n_alphas': [200, 500],
                            'normalize':[False, True]},
                'RSCV'  : {}
                }},

        'Lasso'         : {
            'estimator' : Lasso(random_state=123456),
            'parameters': {
                'GSCV'  : { 'alpha': [0.001, 0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 1, 10],
                            'max_iter': [15000],
                            'normalize':[False, True],
                            'precompute':[False, True]},
                'RSCV'  : {}
                }},

        'LLIC'          : {
            'estimator' : LassoLarsIC(fit_intercept=True,
                                      verbose=False,
                                      normalize=True,
                                      precompute='auto',
                                      max_iter=10000,
                                      eps=2.220446049250313e-16,
                                      copy_X=True,
                                      positive=False),
            'parameters': {
                'GSCV'  : { 'criterion' : ['aic','bic'],
                            },
                'RSCV'  : {}
                }},

        'ENet'          : {
            'estimator' : ElasticNetCV( cv=10,
                                        alphas=[0.001, 0.01, 0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 1, 10],
                                        random_state=123456,
                                        n_jobs= 20,
                                        l1_ratio=[.01,.05, .1, .3, .5, .7, .9, .98]),
            'parameters': {
                'GSCV'  : { 'n_alphas' : [200, 500],
                            'max_iter' : [20000, 50000],
                            'tol'      : [1e-3, 1e-4, 1e-5, ],
                            'normalize': [False, True]},
                'RSCV'  : {}
                }},
        }

    def Regression(self):
        return {
        'RF'   : {
            'estimator' : RandomForestRegressor(n_jobs=-1, oob_score=True, n_estimators=200),
            'parameters': {
                'GSCV'  : { 'max_features': (0.4, 0.6, 0.8, 'auto'),
                            'max_depth': (5, 8, 12, 15, 25, None),
                            'n_estimators': [100, 200, 400, 600],
                            'min_samples_split':[2,5,10] },
                'RSCV'  : {}
                }},

        'Decis_Tree'    : {
            'estimator' : DecisionTreeRegressor(),
            'parameters': {
                'GSCV'  : { 'max_features': (0.4, 0.6, 0.8, 'auto'),
                            'max_depth': (5, 8, 12, 15, 25, None)},
                'RSCV'  : {}
                }},

        'SVM'           : {
            'estimator' : SVR(gamma='scale', kernel="linear"),
            'parameters': {
                'GSCV'  : {'C': [0.1, 0.5, 1, 5, 10, 100, 1000]},
                'RSCV'  : {}
                }},


        'MLP'           : {
            'estimator' : MLPRegressor(max_iter=3000, solver='lbfgs'),
            'parameters': {
                'GSCV'  : { 'alpha': [10**i for i in [-5, -3, -1, 1, 3]],
                            'max_iter':[10000, 15000],
                            'solver':['adam', 'lbfgs'],
                            'hidden_layer_sizes': [(50, 40, 20), (100, 30, 20), (200, 20), (10, 10, 10), (100, 20, 10)],
                            'learning_rate': ["constant", "invscaling", "adaptive"],
                            'activation': ["logistic", "relu", "tanh"]},
                'RSCV'  : {}
                }},

        'LassoCV'       : {
            'estimator' : LassoCV(cv=10, alphas=[0.0005, 0.001, 0.01,0.1, 0.5, 0.9, 1, 10], max_iter=5000),
            'parameters': {
                'GSCV'  : { 'n_alphas': [200, 500],
                            'max_iter':[5000, 10000, 15000],
                            'normalize':[False, True]},
                'RSCV'  : {}
                }},

        'Lasso'         : {
            'estimator' : Lasso(alpha=1, max_iter=5000),
            'parameters': {
                'GSCV'  : { 'alpha': [0.1, 0.5, 0.9, 1, 5, 10],
                            'max_iter': [5000, 10000, 15000],
                            'normalize':[False, True],
                            'precompute':[False, True]},
                'RSCV'  : {}
                }},

        'ENet'          : {
            'estimator' : ElasticNetCV( cv=10,
                                        alphas=[0.0001, 0.0005, 0.001, 0.01, 0.5, 0.1,1, 10],
                                        max_iter=5000,
                                        l1_ratio=[.01, .1, .5, .7, .9, .95, .99]),
            'parameters': {
                'GSCV'  : { 'n_alphas': [200, 500],
                            'max_iter':[5000, 10000, 15000],
                            'normalize':[False, True]},
                'RSCV'  : {}
                }},

        'BayesianRidge' : {
            'estimator' : '',
            'parameters': {
                'GSCV'  : {},
                'RSCV'  : {}
                }},

        'LogistR'       : {
            'estimator' : LogisticRegression(random_state=None,
                                              solver='lbfgs',
                                              fit_intercept=True,
                                              max_iter=500,
                                              multi_class='auto'),
            'parameters': {
                'GSCV'  : {},
                'RSCV'  : {}
                }},
        }




import numpy as np
from scipy.special import expit
from sklearn.model_selection import (GridSearchCV, RandomizedSearchCV, StratifiedShuffleSplit, LeavePOut, 
                                     LeaveOneOut, RepeatedStratifiedKFold, StratifiedKFold, RepeatedKFold)
from sklearn.preprocessing import LabelEncoder, LabelBinarizer
from sklearn.utils.validation import column_or_1d
import warnings
import threading

def _predict_proba_lr(prob):
    """
    Probability estimation for OvR logistic regression.
    Positive class probabilities are computed as
    1. / (1. + np.exp(-self.decision_function(X)));
    multiclass is handled by normalizing that over all classes.
    """
    prob = expit(prob)
    if prob.ndim == 1:
        return np.vstack([1 - prob, prob]).T
    else:
        # OvR normalization, like LibLinear's predict_probability
        prob /= prob.sum(axis=1).reshape((prob.shape[0], -1))
    return prob

def Check_Label(_y):
    y = column_or_1d(_y)
    if y.min() >0:
        warnings.warn("The taget babel should start with 0,, place check you data format.")
        return LabelEncoder().fit_transform(y)
    else:
        return _y

def Check_proba(prob):
    if isinstance(prob, list) or isinstance(prob, tuple):
        prob = np.array(prob)
    if prob.ndim==1:
        prob = np.array([1-prob, prob]).T
        #if isinstance(prob, pd.Series):
    if prob.ndim==2:
        if prob.shape[1] ==1:
            prob = np.c_[1-prob, prob]
        return np.array(prob)
    else:
        raise ValueError('The data type is unkown.')

def Check_Binar(_y):
    y = column_or_1d(_y)
    y = LabelBinarizer().fit_transform(y)
    if y.shape[1] ==1:
        y = np.c_[1-y,y]
    return y

def CrossvalidationSplit(CVt='SSA', n_splits=10, n_repeats=3, leavP=1, test_size=0.4, random_state = None):
    n_splits, n_repeats, leavP = int(n_splits), int(n_repeats), int(leavP)
    CV = { 'SSS' : StratifiedShuffleSplit(
                        n_splits = n_splits,
                        test_size= test_size,
                        random_state=random_state),
            'SSA': StratifiedShuffleSplit(
                        n_splits = n_splits,
                        test_size= test_size,
                        random_state=random_state),
            'SFA': StratifiedKFold(
                        n_splits=int(round(1/test_size)),
                        shuffle=True,
                        random_state=random_state),
            'SFK': StratifiedKFold(
                        n_splits = n_splits,
                        shuffle=True,
                        random_state=random_state),
            'RKF': RepeatedKFold(
                        n_splits = n_splits,
                        n_repeats= n_repeats,
                        random_state=random_state),
            'RSKF': RepeatedStratifiedKFold(
                        n_splits = n_splits,
                        n_repeats= n_repeats,
                        random_state=random_state),
            'LOU': LeaveOneOut(),
            'LPO': LeavePOut(leavP),
    }
    return CV[CVt]

class MyThread(threading.Thread):
    def __init__(self, func, args=()):
        super(MyThread, self).__init__()
        self.func = func
        self.args = args
    def run(self):
        self.result = self.func(*self.args)
    def get_result(self):
        try:
            return self.result
        except Exception:
            return None


#===================================================
#
# Project      : mlsuite building models or predicting
# Description  : pipeline script
# Usage        : 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-06-14
#
#====================================================


from .MLSupervising import run_fitting
from .MLPredicting import run_predict

class Pipeline():
    'The pipeline used for machine learning models'
    def __init__(self, arg, log,  *array, **dicts):
        self.arg = arg
        self.log = log
        self.array  = array
        self.dicts  = dicts

    def Pipe(self):

        if self.arg.commands in ['Fitting', 'Auto']:
            run_fitting(self.arg, self.log)

        if self.arg.commands in ['Predict', 'Auto']:
            run_predict(self.arg, self.log)



#===================================================
#
# Project      :mlsuite building models or predicting
# Description  : main script
# Usage        : 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-06-21
#
#====================================================


import traceback
import os,sys
import time
from .MLPipeline import Pipeline
from .MLLogging import Logger
from .MLArguments import Args


def Commands():
    info = '''
    #===================================================
    #
    # Project      : mlsuite building models or predicting
    # Description  : main script
    # Usage        : 
    # Author       : Suxing Li
    # Email        : li.suxing@genecast.com.cn
    # Created at   : 2022-06-21
    #
    #====================================================
    '''
    
    args = Args()
    args.header = '/%s_' % (args.model)
    args.outdir = '%s/%s/' % (args.outdir, args.model)
    args.output = '%s%s' % (args.outdir, args.header)
    args.MODELF = '%s/00MODEL/%s' % (args.outdir, args.header)
    Theader = time.strftime("%Y%m%d%H%M", time.localtime())
    if args.commands=='Predict' :
        if ( bool(args.out_predict) | bool(args.modelpath) ):
            args.outdir = '%s/%s/' % (args.modelpath, args.model)
            args.output = '%s/%s_Prediction_%s/%s' % (args.out_predict, args.out_header, Theader, args.header)
            args.MODELF = '%s/00MODEL/%s' % (args.outdir, args.header)
    os.makedirs(os.path.dirname(args.output), exist_ok=True)

    Log = Logger( '%s%s_log.log'%(args.output, args.commands) )
    Log.NIF(info.strip())
    Log.NIF("The argument you have set as follows:".center(59, '*'))
    for i,k in enumerate(vars(args),start=1):
        Log.NIF('**%s|%-13s: %s'%(str(i).zfill(2), k, str(getattr(args, k))) )
    Log.NIF(59 * '*')

    return_code = 0
    try:
        Pipeline(args, Log).Pipe()
        Log.CIF('Success!!!')
    except Exception:
        traceback.print_exc()
        return_code = 1
    finally:
        Log.CIF('You can check your progress in log file.')
        exit(return_code)




#!/usr/bin/env python
# -*- coding:utf-8 -*-

from setuptools import find_packages, setup
from glob import glob
from os.path import dirname, join
import os

setup(
    name='mlsuite',
    version='2.1.7',
    description='The traditional machine learning analysis based on sklearn package',
    author='suxing li',
    author_email='li.suxing@genecast.com.cn',
    maintainer='suxing li',
    maintainer_email='li.suxing@genecast.com.cn',
    packages=find_packages(where='.', exclude=(), include=('*',)),
    include_package_data=True,
    platforms=['all'],
    url='https://git.genecast.com.cn/narwhal/mlsuite',
    scripts=['./mlsuite.py'],
    install_requires=[
        'lightgbm == 3.3.3',
        'joblib == 1.2.0',
        'numpy == 1.21.4',
        'pandas == 1.1.5',
        'scikit-learn == 1.2.0',
        'sklearn_pandas == 2.2.0',
        'xgboost == 1.7.2'
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Healthcare Industry",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
	"Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Visualization",
    ]
)



import pandas as pd
import numpy as np
import re
import os

class Openf():
    def __init__(self, file, *array,compression='infer',comment='#', index_label=None, sep="\t",index = True, header=0,index_col=None):
        self.file  = file.strip()
        self.array = array
        self.compr = compression
        self.comme = comment
        self.sep   = sep
        self.header= header
        self.index_col= index_col
        self.index = index 
        self.index_label=index_label

    def openb(self):
        if re.search(r'.*xlsx$|.*xlss$', self.file, re.I):
            outp = pd.read_excel(self.file, header=self.header, index_col=self.index_col, encoding='utf-8', sheet_name=0, comment=self.comme, compression=self.compr).fillna(np.nan)
        else:
            outp = pd.read_csv(self.file, header=self.header, index_col=self.index_col, encoding='utf-8', sep=self.sep, comment=self.comme,compression=self.compr).fillna(np.nan)
        return(pd.DataFrame(outp))

    def openw(self):
        os.makedirs( os.path.dirname(self.file), exist_ok=True )
        f= open(self.file, 'w',encoding='utf-8')
        f.write(self.array[0])
        f.close()
    def openv(self):
        os.makedirs( os.path.dirname(self.file), exist_ok=True )
        df = self.array[0]
        df.to_csv(self.file, sep=self.sep, index=self.index, index_label=self.index_label, header=True)

class OpenM():
    def __init__(self, arg, log):
        self.arg = arg
        self.log = log

    def openi(self):
        dfall = Openf(self.arg.input, index_col=0).openb()
        return (dfall)
    def openg(self):
        group = Openf(self.arg.group).openb()
        group.columns = ['Variables', 'Group', 'Type']

        RYa = group[(group.Group == 'Y') & (group.Type =='R') ].Variables.tolist()
        CYa = group[(group.Group == 'Y') & (group.Type =='C') ].Variables.tolist()
        Xa  = group[(group.Group != 'Y')].Variables.tolist()
        Xg  = group[(group.Group != 'Y')][['Variables','Group','Type']]
        Xg.set_index('Variables', inplace=True)

        return (group, RYa, CYa, Xa, Xg)

    def openp(self):
        dfpred = Openf(self.arg.predict, index_col=0).openb()
        return (dfpred)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#===================================================
#
# Project      : mlsuite building models or predicting
# Description  : main script
# Usage        : 
# Author       : Suxing Li
# Email        : li.suxing@genecast.com.cn
# Created at   : 2022-06-21
#
#====================================================


from MLsuite import main

if __name__ == "__main__":
    main.Commands()
